"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.bareLength = exports.combine = exports.bleach = exports.interpolate = exports.parseGradient = exports.parseMethod = exports.repeatChar = exports.ucFirst = exports.formatTime = exports.toFixed = void 0;
var constants_1 = require("./constants");
function toFixed(value, precision) {
    var power = Math.pow(10, precision);
    return (Math.round(value * power) / power).toFixed(precision);
}
exports.toFixed = toFixed;
function formatTime(ms) {
    return Number.isNaN(ms) || !Number.isFinite(ms)
        ? '0.0'
        : toFixed(ms / 1000, 1);
}
exports.formatTime = formatTime;
function lcFirst(str) {
    return str.charAt(0).toLowerCase() + str.substring(1);
}
function ucFirst(str) {
    return str.charAt(0).toUpperCase() + str.substring(1);
}
exports.ucFirst = ucFirst;
function repeatChar(char, count) {
    return new Array(count + 1).join(char);
}
exports.repeatChar = repeatChar;
function parseMethod(cursor, str) {
    var input = str.substring(1);
    return (parseColor(cursor, input) ||
        parseFont(cursor, input) ||
        parseGradient(input));
}
exports.parseMethod = parseMethod;
function parseColor(cursor, str) {
    var match = str.match(/^(bgR|r)ed/) ||
        str.match(/^(bgB|b)lue/) ||
        str.match(/^(bgC|c)yan/) ||
        str.match(/^(bgG|g)rey/) ||
        str.match(/^(bgW|w)hite/) ||
        str.match(/^(bgB|b)lack/) ||
        str.match(/^(bgG|g)reen/) ||
        str.match(/^(bgY|y)ellow/) ||
        str.match(/^(bgM|m)agenta/) ||
        str.match(/^(bgB|b)right(Black|Red|Green|Yellow|Blue|Magenta|Cyan|White)/);
    if (match) {
        var method = match[0];
        var suffix = str.substring(method.length);
        var isBg = method.startsWith('bg');
        if (isBg) {
            method = lcFirst(method.substring(2));
        }
        if (typeof cursor[method] === 'function') {
            return {
                isBg: isBg,
                method: method,
                suffix: suffix,
            };
        }
    }
}
function parseFont(cursor, str) {
    var match = str.match(/^bold|italic|underline|inverse/);
    if (match) {
        var method = match[0];
        var suffix = str.substring(method.length);
        if (typeof cursor[method] === 'function') {
            return {
                isFont: true,
                method: method,
                suffix: suffix,
            };
        }
    }
}
function parseGradient(str) {
    var match = str.match(/^\((.+),(.+)\)/);
    if (match) {
        var m1 = match[1].trim();
        var m2 = match[2].trim();
        var color1 = m1.startsWith('#') ? hex2rgb(m1) : name2rgb(m1);
        var color2 = m2.startsWith('#') ? hex2rgb(m2) : name2rgb(m2);
        if (color1 && color2) {
            return {
                method: 'gradient',
                color1: color1,
                color2: color2,
            };
        }
    }
}
exports.parseGradient = parseGradient;
function interpolate(color1, color2, percent) {
    return {
        r: atPercent(color1.r, color2.r, percent),
        g: atPercent(color1.g, color2.g, percent),
        b: atPercent(color1.b, color2.b, percent),
    };
}
exports.interpolate = interpolate;
function atPercent(a, b, percent) {
    return a + Math.round((b - a) * percent);
}
function hex2rgb(color) {
    var c = color.substring(1);
    var r = c.substring(0, 2);
    var g = c.substring(2, 4);
    var b = c.substring(4, 6);
    return {
        r: parseInt(r, 16),
        g: parseInt(g, 16),
        b: parseInt(b, 16),
    };
}
function name2rgb(name) {
    var hex = {
        red: '#ff0000',
        blue: '#0000ff',
        cyan: '#00ffff',
        grey: '#808080',
        white: '#ffffff',
        black: '#000000',
        green: '#008000',
        yellow: '#ffff00',
        magenta: '#ff00ff',
    }[name];
    return hex ? hex2rgb(hex) : null;
}
function bleach(output) {
    return (output
        .replace(/\.(bgR|r)ed/g, '')
        .replace(/\.(bgB|b)lue/g, '')
        .replace(/\.(bgC|c)yan/g, '')
        .replace(/\.(bgG|g)rey/g, '')
        .replace(/\.(bgW|w)hite/g, '')
        .replace(/\.(bgB|b)lack/g, '')
        .replace(/\.(bgG|g)reen/g, '')
        .replace(/\.(bgY|y)ellow/g, '')
        .replace(/\.(bgM|m)agenta/g, '')
        // bright
        .replace(/\.(bgB|b)right(Black|Red|Green|Yellow|Blue|Magenta|Cyan|White)/g, '')
        // font style
        .replace(/\.bold|italic|underline|inverse/g, '')
        // gradient
        .replace(/\.gradient\((.+),(.+)\)/g, ''));
}
exports.bleach = bleach;
function combine(output, filled, blank, bare) {
    var bar = filled + blank;
    if (!bare) {
        bar = bar || constants_1.placeholder;
        // eslint-disable-next-line no-param-reassign
        blank = blank || constants_1.placeholder;
        // eslint-disable-next-line no-param-reassign
        filled = filled || constants_1.placeholder;
    }
    return output
        .replace(/:filled/g, filled)
        .replace(/:blank/g, blank)
        .replace(/:bar/g, bar);
}
exports.combine = combine;
function bareLength(output) {
    var str = output
        .replace(/:filled/g, '')
        .replace(/:blank/g, '')
        .replace(/:bar/g, '');
    return str.length;
}
exports.bareLength = bareLength;
//# sourceMappingURL=util.js.map